/*
 * cp40_hfile.c --
 *
 *	Implementation of simple editable hash-table files.  Another
 *	improvement with respect to actual implementation of management of
 *	simple hash table is that I tried to do the necessary housekeeping
 *	to avoid problems if some of these routines get interrupted,
 *	e.g. by user typing ^C.  This may seem too paranoid but is really
 *	needed if the CP40 library is linked with some interactive language,
 *	e.g. Yorick or IDL, which allow such interruptions.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 1995-2006 Eric Thibaut.
 *
 *	This file is part of CP40soft.
 *
 *	CP40soft is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	CP40soft is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with CP40soft; if not, write to the Free Software Foundation,
 *	Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */

#include <unistd.h>
#include <stdlib.h>

#include "cp40.h"

/* Private routines and macros. */

/* Private data. */
static const char *empty= "";
static FILE *db_file= NULL; /* Database file stream. */

/*_____________________________________________________________________________
 *
 * cp40_hash_restore --
 *
 *	Read editable key-value file and store its contents in a hash table.
 *
 * Return:
 *	Standard result: CP40_OK or CP40_ERROR.
 *
 * Side effects:
 *	Output cp40_hash_table_t is assumed to have been correctly initialized
 *	by cp40_hash_init_table(table, CP40_HASH_STRING_KEYS) and its contents may
 *	be overwriten.
 *_____________________________________________________________________________
 */

int cp40_hash_restore(cp40_hash_table_t *table, const char *filename)
{
  static cp40_dstr_t *linePtr= NULL;
  static cp40_dstr_t *keyPtr= NULL;
  static cp40_dstr_t *valuePtr= NULL;
  static int init= 0;
  char *ptr, *string;
  int i, c, status= CP40_OK, eof=0, new;
  int length, line_count= 0;
  cp40_hash_entry_t *entry;

  if (! init) {
#undef  CANNOT_ALLOC
#define CANNOT_ALLOC(ptr, type) (ptr==NULL && (ptr= CP40_ALLOC_ONE(type))==NULL)
    CP40_ASSERT(! CP40_IS_SPACE('\0')); /* The following code relies on the fact
				     that '\0' is not a space. */
    if (   CANNOT_ALLOC(linePtr, cp40_dstr_t)
	|| CANNOT_ALLOC(keyPtr, cp40_dstr_t)
	|| CANNOT_ALLOC(valuePtr, cp40_dstr_t)) {
      cp40_sys_error("cannot create hash table strings");
      goto error;
    }
    cp40_dstr_init(linePtr);
    cp40_dstr_init(keyPtr);
    cp40_dstr_init(valuePtr);
    init= 1;
  } else {
    cp40_dstr_reset(linePtr);
    cp40_dstr_reset(keyPtr);
    cp40_dstr_reset(valuePtr);
  }
  if (table->key_type != CP40_HASH_STRING_KEYS) {
    return cp40_error("hash table key type must be `CP40_HASH_STRING_KEYS`");
  }
  if (cp40_open_static_file(&db_file, filename, "r") != CP40_OK) {
    return CP40_ERROR;
  }

  for (;;) {
    /* Read next line and get first non-space character. */
    if (! eof && cp40_dstr_read_line(linePtr, db_file, &eof, &line_count)!=CP40_OK) {
      goto error;
    }
    for (ptr=cp40_dstr_value(linePtr) ; CP40_IS_SPACE(ptr[0]) ; ptr++)
      ;
    c= ptr[0];

    /* Possibly append continuation line (append a newline if continuation
       line is filled with spaces). */
    if (c == '>') {
      if (cp40_dstr_length(keyPtr) <= 0) {
	/* Empty key string: nowhere to append anything... */
	goto unexpected_char;
      }
      for (ptr+=1, i=0 ; CP40_IS_SPACE(ptr[i]) ; i++)
	;
      if ((ptr[i]==0 ? cp40_dstr_append_char(valuePtr, '\n')
	   : cp40_dstr_append(valuePtr, ptr, -1)) != CP40_OK) {
	goto error;
      }
      continue;
    }

    /* Save previously parsed key-value pair if any and reset the
       corresponding dynamic strings to be empty. */
    if (cp40_dstr_length(keyPtr) > 0) {
      entry= cp40_hash_create_entry(table,cp40_dstr_value(keyPtr), &new);
      if (entry == NULL || (string= cp40_dstr_steal(valuePtr)) == NULL) {
	goto error;
      }
      cp40_hash_set_value(entry, string, CP40_HASH_DYNAMIC);
      cp40_dstr_reset(keyPtr);
    }

    /* Skip empty line and comments. */
    if (c == 0 || c == '#') {
      if (eof) break;
      continue;
    }

    /* Store key and value strings. */
    for (length=i=0 ; ; i++) {
      if ((c= ptr[i]) == '=') {
	length= i;
	break;
      }
      if (CP40_IS_SPACE(c)) {
	length= i;
	for (i+=1 ; CP40_IS_SPACE(ptr[i]) ; i++)
	  ;
	c= ptr[i];
	break;
      }
      if (c==0 || c=='>') {
	break;
      }
    }
    if (length <= 0 || c != '=') {
      goto unexpected_char;
    }
    if (cp40_dstr_append(keyPtr, ptr, length) != CP40_OK) {
      goto error;
    }
    for (ptr+=i+1 ; CP40_IS_SPACE(ptr[0]) ; ptr++)
      ;
    if (ptr[0] != 0 && cp40_dstr_append(valuePtr, ptr, -1) != CP40_OK) {
      goto error;
    }
  }

 done:
  cp40_close_static_file(&db_file);
  cp40_dstr_free(linePtr);
  cp40_dstr_free(keyPtr);
  cp40_dstr_free(valuePtr);
  return status;

 unexpected_char:
  if (c >= ' ' && c <= '~') {
    cp40_error("unexpected char '%c' in hash file at line %d", c, line_count);
  } else if (c==0) {
    cp40_error("unexpected end-of-line in hash file at line %d", line_count);
  } else {
    cp40_error("unexpected char 0x%02x in hash file at line %d", (unsigned)c,
	    line_count);
  }
 error:
  status= CP40_ERROR;
  goto done;
}

/*_____________________________________________________________________________
 *
 * cp40_hash_save --
 *
 *	Simple driver to save hash table into editable file.
 *
 * Return:
 *	Standard result: CP40_OK or CP40_ERROR.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

int cp40_hash_save(cp40_hash_table_t *table, const char *filename, int overwrite)
{
  cp40_hash_entry_t *entry;
  cp40_hash_search_t search;

  if (table->key_type != CP40_HASH_STRING_KEYS) {
    return cp40_error("hash table key type must be `CP40_HASH_STRING_KEYS'");
  }
  if (filename == NULL || filename[0] == 0) {
    return cp40_error("bad file name for saving hash-table");
  }
  if (access(filename, F_OK) == 0 && ! overwrite) {
    return cp40_error("cannot save hash-table: file already exists");
  }
  if (cp40_open_static_file(&db_file, filename, "w") != CP40_OK) {
    return CP40_ERROR;
  }

  /* Loops over every hash entries. */
  for (entry= cp40_hash_first_entry(table, &search);
       entry != NULL;
       entry= cp40_hash_next_entry(&search)) {
    if (cp40_hash_write_entry(db_file, entry->key.string,
		       (char *)entry->client_data, 0) != CP40_OK) {
      cp40_close_static_file(&db_file);
      return CP40_ERROR;
    }
  }
  cp40_close_static_file(&db_file);
  return CP40_OK;
}

/*_____________________________________________________________________________
 *
 * cp40_hash_write_comment --
 *
 *	Print out a comment line in a database file (STRING must not contain
 *	any newline; this is not checked).
 *
 * Return:
 *	None.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

int cp40_hash_write_comment(FILE *file, const char *string)
{
  if ((string!=NULL && (fputs("# ", file) < 0 || fputs(string, file) < 0))
      || putc('\n', file) < 0) {
    return cp40_sys_error("error writing database comment");
  }
  return CP40_OK;
}

/*_____________________________________________________________________________
 *
 * cp40_hash_write_entry --
 *
 *	Print a hash entry key-value pair taking care of splitting value
 *	if it is too long (if COMMENT is non-zero the resulting line(s) get
 *	commented).
 *
 * Return:
 *	None.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

#define MAX_LENGTH 50

int cp40_hash_write_entry(FILE *file, const char *key, const char *value, int comment)
{
  int c;
  int print= 0;
  char *sign= "= "; /* Equal sign followed by a space. */
  long i0= 0;       /* Begin of string to print. */
  long i1= -1;      /* Current position in value string. */
  long j= -1;       /* Position of last space. */
  const char *format= comment ? "# %18s %s" : "%20s %s";

  if (value==NULL) value= empty;
  if (comment) {
    if (key==NULL) {
      key= empty;
    }
  } else {
    /* Check key. */
    int i= -1;
    if (key!=NULL) {
      while ((c= key[++i]) != 0) {
	if (CP40_IS_SPACE(c) || c=='=' || c=='>') {
	  i= 0;
	  break;
	}
      }
    }
    if (i<=0) {
      cp40_warning("skipping hash table entry (bad key)");
      return CP40_OK;
    }
  }

  /* Write hash entry key-value. */
  for (;;) {
    c= value[++i1];
    if (c==0 || CP40_IS_SPACE(c)) {
      j= i1;
      print= c==0 || c=='\n';
    }
    if (print || (i1 - i0) >= MAX_LENGTH) {
      if (j <= i0) {
	j= i1;
      }
      if (j >= i0) {
	if (fprintf(file, format, key, sign) <= 0
	    || (j > i0 && fwrite(value+i0, 1, j-i0, file) != j-i0)
	    || putc('\n', file) < 0) {
	  break;
	}
      }
      if ((c= value[j]) == 0) {
	return CP40_OK;
      }
      if (c == '\n') {
	if (fprintf(file, format, empty, ">\n") <= 0) {
	  break;
	}
	i0= j+1;
      } else {
	i0= j;
      }
      key= empty;
      sign= ">"; /* Continuation sign followed by NO space. */
      print= 0;
    }
  }
  return cp40_sys_error("error writing hash table entry");
}
/*---------------------------------------------------------------------------*/
