/*
 * cp40_high.c --
 *
 *      Implementation of the interface to manage CP40 data instance
 *      and high level attributes (i.e. by string keys and values).
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 1995-2006 Eric Thibaut.
 *
 *	This file is part of CP40soft.
 *
 *	CP40soft is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	CP40soft is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with CP40soft; if not, write to the Free Software Foundation,
 *	Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "cp40.h"

/*---------------------------------------------------------------------------*/
/* PUBLIC ROUTINES PROVIDED BY THIS FILE */

extern int cp40_save_attr(cp40_instance_t *instance, const char *filename,
			  int flags);
extern int cp40_restore_attr(cp40_instance_t *instance, const char *filename,
			     int flags);
extern cp40_instance_t *cp40_new_instance(void);
extern int cp40_delete_instance(cp40_instance_t *instance, int update);

/*---------------------------------------------------------------------------*/
/* Some attributes are passed by litteral keyword and make use of the
   following definitions. */

typedef struct keyword_info_s keyword_info_t;

struct keyword_info_s {
  const char *sval;  /* string value of keyword */
  const int ival;    /* integer identifier of keyword value */
};

static keyword_info_t keyword_parity[]= {
  {"any", CP40_PARITY_ANY},
  {"even", CP40_PARITY_EVEN},
  {"odd", CP40_PARITY_ODD},
  {NULL, 0},
};

static keyword_info_t keyword_spurious[]= {
  {"leave", CP40_SPURIOUS_LEAVE},
  {"delete", CP40_SPURIOUS_DELETE},
  {"centre", CP40_SPURIOUS_CENTRE},
  {"single", CP40_SPURIOUS_SINGLE},
  {NULL, 0},
};

static keyword_info_t keyword_correction[]= {
  {"none", CP40_CORRECTION_NONE},
  {"simple", CP40_CORRECTION_SIMPLE},
  {"random", CP40_CORRECTION_RANDOM},
  {"weighted",CP40_CORRECTION_WEIGHTED},
  {NULL, 0},
};

static keyword_info_t keyword_pixel[]= {
  {"1/4", CP40_PIXEL_LOGICAL},
  {"1/2", CP40_PIXEL_HALF_PHYSICAL},
  {"1", CP40_PIXEL_PHYSICAL},
  {NULL, 0},
};

#if 0
static keyword_info_t keyword_mode[]= {
  {"none", CP40_IO_MODE_NONE},
  {"read", CP40_IO_MODE_READ},
  {"write", CP40_IO_MODE_WRITE},
  {"update", CP40_IO_MODE_UPDATE},
  {NULL, 0},
};
#endif

/*---------------------------------------------------------------------------*/
/* Definitions for attributes handlers. */

typedef struct attr_info_s attr_info_t;

typedef int attr_get_op_t(cp40_instance_t *instance, const char *name,
			  cp40_dstr_t *value);
typedef int attr_set_op_t(cp40_instance_t *instance, const char *name,
			  const char *value);

struct attr_info_s {
  const char *name;
#undef  IGNORED
#define IGNORED        (0)
#undef  READ_ONLY
#define READ_ONLY      (1<<0)
#undef  PER_CHANNEL
#define PER_CHANNEL    (1<<1)
#undef  PREPROCESSING
#define PREPROCESSING  (1<<2)
#undef  EXTRA
#define EXTRA          (1<<3)
  int flags;
  attr_get_op_t *get;
  attr_set_op_t *set;
};

static attr_get_op_t get_nstaturations, get_nframes, get_frame_date;
static attr_get_op_t get_nphotons, get_frame_count;
static attr_get_op_t get_exposure;
static attr_set_op_t set_exposure;
static attr_get_op_t get_extra, get_scale, get_angle, get_coefs;
static attr_set_op_t set_extra, set_scale, set_angle, set_coefs;
static attr_get_op_t get_range, get_window;
static attr_set_op_t set_range, set_window;
static attr_get_op_t get_correction, get_parity, get_pixel;
static attr_get_op_t get_spurious;
static attr_set_op_t set_correction, set_parity, set_pixel;
static attr_set_op_t set_spurious;
static attr_set_op_t set_bogus, set_read_only;
static attr_get_op_t get_bogus;

static attr_info_t std_attr[]= {
  /* global file info: */
  {"nframes", READ_ONLY, get_nframes, set_read_only},
  {"nphotons", PER_CHANNEL | READ_ONLY, get_nphotons, set_read_only},
  {"nsaturations", READ_ONLY, get_nstaturations, set_read_only},

  /* current frame info: */
  {"frame_count", IGNORED, get_frame_count, set_read_only},
  {"frame_date", IGNORED, get_frame_date, set_read_only},

  /* required/effective supports: */
  {"window", PER_CHANNEL | PREPROCESSING, get_window, set_window},
  {"extreme", PER_CHANNEL | READ_ONLY, get_window, set_read_only},

  /* preprocessing */
  {"exposure", PREPROCESSING, get_exposure, set_exposure},
  {"spurious", PREPROCESSING, get_spurious, set_spurious},
  {"pixel", PREPROCESSING, get_pixel, set_pixel},
  {"correction", PREPROCESSING, get_correction, set_correction},
  {"scale", PREPROCESSING, get_scale, set_scale},
  {"angle", PREPROCESSING, get_angle, set_angle},
  {"raw_coefs", PREPROCESSING, get_coefs, set_coefs},
  {"coefs", IGNORED, get_coefs, set_read_only},

  /* frame selection rules */
  {"parity", PREPROCESSING, get_parity, set_parity},
  {"flux", PER_CHANNEL | PREPROCESSING, get_range, set_range},
  {"spread", PER_CHANNEL | PREPROCESSING, get_range, set_range},
  {"distance", PER_CHANNEL | PREPROCESSING, get_range, set_range},
  {"ratio", PER_CHANNEL | PREPROCESSING, get_range, set_range},
  {NULL, 0, NULL, NULL}
};

static attr_info_t attr_bogus= {NULL, 0, get_bogus, set_bogus};
static attr_info_t attr_extra= {NULL, 0, get_extra, set_extra};
static FILE *attr_file= NULL;

/*---------------------------------------------------------------------------*/
/* PRIVATE MACROS AND DATA */

/* Loop over each CP40 channels. */
#undef EACH_CHANNEL
#define EACH_CHANNEL(var) (var=0 ; var<CP40_NCHANNELS ; var++)

/* The maximum length of a string representing a double precision floating
   point value is 24 characters: 1 sign + 17 significant digits + 1 dot
   + 1 'e' or 'E' + 1 exponent sign + 3 exponent digits. */
#undef DOUBLE_SPACE
#define DOUBLE_SPACE 25

#define ATTR_ALL_CHANNELS (-1)

/*---------------------------------------------------------------------------*/
/* DYNAMIC CACHE BUFFER TO TEMPORARILY STORE INPUT STRING TO PARSE. */

static char *clone_in_cache(const char *string);

static char *clone_in_cache(const char *string)
{
  static cp40_dstr_t cache;
  static int need_init= 1;

  if (need_init) {
    cp40_dstr_init(&cache);
    need_init= 0;
  } else {
    cp40_dstr_reset(&cache);
  }
  if (cp40_dstr_append(&cache, string, -1) != CP40_OK) return NULL;
  return cp40_dstr_value(&cache);
}

/*---------------------------------------------------------------------------*/

static int parse_channel(int *channelPtr, const char *name);
static attr_info_t *get_attr_info(const char *name);

/*_____________________________________________________________________________
 *
 * cp40_new_instance --
 *
 *	Make a new CP40 data instance: allocate a new cp40_instance_t structure
 *	and set it with default attributes.
 *
 * Return:
 *	A pointer to a new cp40_instance_t structure; NULL in case of error
 *	(error message can be recovered by cp40_error_info).
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

cp40_instance_t *cp40_new_instance(void)
{
  int channel;
  cp40_window_t window;
  cp40_range_t range;
  cp40_instance_t *instance;

  window.xmin = CP40_WINDOW_WILD;
  window.xmax = CP40_WINDOW_WILD;
  window.ymin = CP40_WINDOW_WILD;
  window.ymax = CP40_WINDOW_WILD;

  range.flags = 0;
  range.min = 0.0;
  range.max = 0.0;

  /* Allocate memory for cp40_instance_t structure and perform very preliminary
     initialization. */
  if (! CP40_ALLOC_AT(instance, 1)) {
    (void)cp40_sys_error("cannot allocate cp40_instance_t structure");
    return NULL;
  }
  (void)memset(instance, 0, sizeof *instance);
  cp40_init_stream(&instance->stream);
  cp40_hash_init_table(&(instance->extra_table), CP40_HASH_STRING_KEYS);

  /* Higher level initialization. */
  if (cp40_set_exposure(instance, 1) != CP40_OK) {
    cp40_mem_free(instance);
    return NULL;
  }
  cp40_set_parity(instance, CP40_PARITY_ANY);
  cp40_set_pixel(instance, CP40_PIXEL_PHYSICAL);
  cp40_set_spurious(instance, CP40_SPURIOUS_LEAVE);
  cp40_set_correction(instance, CP40_CORRECTION_NONE);
  for EACH_CHANNEL(channel) {
    cp40_set_window(instance, channel, &window);
    cp40_set_distance(instance, channel, &range);
    cp40_set_flux(instance, channel, &range);
    cp40_set_ratio(instance, channel, &range);
    cp40_set_spread(instance, channel, &range);
  }

  /* Allocate raw photon buffers. */
  if (! (instance->raw_frame= cp40_new_raw_frame())) {
    cp40_delete_instance(instance, 0);
    return NULL;
  }
  for EACH_CHANNEL(channel) {
    _cp40_init_frame(&instance->frame[channel], instance, channel);
  }
  return instance;
}

/*_____________________________________________________________________________
 *
 * cp40_delete_instance --
 *
 *	Delete a CP40 data instance: free all resources related to the CP40
 *	data instance, close associated stream (if any), delete client
 *	data, ...
 *
 * Arguments:
 *	cp40_instance_t *instance	CP40 data instance.
 *	int        update       If non-zero and CP40 data instance has an
 *				associated stream in mode write/update and
 *				is seekable, rewrite header before closing.
 *
 * Return:
 *	Standard result: CP40_OK or CP40_ERROR (error message can be
 *	recovered by cp40_error_info).
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

int cp40_delete_instance(cp40_instance_t *instance, int update)
{
  cp40_extra_t *extra_list;
  int index, channel, result= CP40_OK;

  if (instance == NULL) return CP40_ERROR;

  /* Cancel all tasks and real-time statistics. */
  cp40_cancel_all_tasks(instance);
  cp40_set_stat_duration(instance, 0);

  /* Free client data. */
  if (instance->delete_client_data) {
    instance->delete_client_data(instance->client_data);
  }

  /* Free distortion correction resources. */
  _cp40_invalidate_cache(instance);

  /* Maybe update header and close associated stream if any. */
  if (instance->stream.ops->mode != CP40_IO_MODE_NONE
      && cp40_close_stream(instance, update) != CP40_OK) {
    result= CP40_ERROR;
  }

  /* Free hash table and list of extra attributes. */
  if ((extra_list = instance->extra_list) != NULL) {
    for (index = 0; index < instance->extra_count; index++) {
      if (extra_list[index].string != NULL) {
	cp40_mem_free((void *)extra_list[index].string);
      }
    }
    cp40_mem_free((void *)extra_list);
  }
  cp40_hash_delete_table(&instance->extra_table);

  /* Free internal buffers and the structure itself. */
  cp40_delete_raw_frame(instance->raw_frame);
  for EACH_CHANNEL(channel) {
    instance->frame[channel].count= 0;
    _cp40_resize_frame(&instance->frame[channel], 0);
  }
  CP40_FREE_MAYBE(instance->count_list);

  /* Free the structure itself. */
  cp40_mem_free((void *)instance);
  return result;
}
/*---------------------------------------------------------------------------*/

static attr_info_t *get_attr_info(const char *name);
/*	Get attribute handler given the attribute name.  Return pointer to
	attribute handler or &attr_bogus in case of error.  At first call,
	initialize internal hash table. */

int cp40_get(cp40_instance_t *instance, const char *name, cp40_dstr_t *value)
{ return get_attr_info(name)->get(instance, name, cp40_dstr_reset(value)); }

int cp40_set(cp40_instance_t *instance, const char *name,
		  const char *value)
{ return get_attr_info(name)->set(instance, name, value); }


static attr_info_t *get_attr_info(const char *name)
{
  static cp40_hash_table_t table;
  static int need_init= 1;
  cp40_hash_entry_t *entry;
  int i, new, chn;

  /* If not yet done, initialize hash table of predefined (standard)
     attributes. */
  if (need_init) {
    cp40_hash_init_table(&table, CP40_HASH_STRING_KEYS);
    for (i = 0; std_attr[i].name; i++) {
      entry= cp40_hash_create_entry(&table, std_attr[i].name, &new);
      if (!entry || !new) {
	/* Unable to initialize hash table for standard attributes
	   or already existing hash entry. */
      bad_init:
	cp40_hash_delete_table(&table);
	return &attr_bogus;
      }
      cp40_hash_set_value(entry, &std_attr[i], CP40_HASH_STATIC);
      if (std_attr[i].flags & PER_CHANNEL) {
	/* Add one entry per channel. */
	char fullname[32]; /* large enough for standard attribute + channel */
	for EACH_CHANNEL(chn) {
	  sprintf(fullname, "%s,%d", std_attr[i].name, chn);
	  entry= cp40_hash_create_entry(&table, fullname, &new);
	  if (!entry || !new) {
	    goto bad_init;
	  }
	  cp40_hash_set_value(entry, &std_attr[i], CP40_HASH_STATIC);
	}
      }
    }
    need_init= 0;
  }

  /* Search attribute in table of predefined (standard) attributes. */
  if ((entry= cp40_hash_find_entry(&table, name)) == NULL) {
    return &attr_extra;
  }
  return (attr_info_t *)cp40_hash_get_value(entry);
}

/*---------------------------------------------------------------------------*/

extern int cp40_save_attr(cp40_instance_t *instance, const char *filename,
			  int flags);
/*	Save CP40 instance attribute into editable attribute file.  Return
	standard result: CP40_OK or CP40_ERROR.  FLAGS are any or'ed combination
	of CP40_ATTR_STANDARD, CP40_ATTR_PREPROCESSING, CP40_ATTR_EXTRA for
	selecting which kind of attributes to save. */

extern int cp40_restore_attr(cp40_instance_t *instance, const char *filename,
			     int flags);
/*	Restore attributes of INSTANCE from editable attribute file FILENAME.
	Return standard result: CP40_OK or CP40_ERROR. FLAGS are any or'ed
	combination of CP40_ATTR_STANDARD, CP40_ATTR_PREPROCESSING,
	CP40_ATTR_EXTRA for selecting which kind of attributes to restore. */

static int print_attr(FILE * file, cp40_instance_t *instance, const char *name,
		      int flags);
/*	Print an attribute-value pair (or one pair per channel) int FILE
	according to given flags and return standard result: CP40_OK or
	CP40_ERROR. */

int cp40_save_attr(cp40_instance_t *instance, const char *filename, int flags)
{
  long i;
  int result = CP40_OK;

  if (filename == NULL || filename[0] == 0) {
    return cp40_error("bad attribute file name");
  }
  if (access(filename, F_OK) == 0) {
    return cp40_error("cannot write attribute: file already exists");
  }
  if (cp40_open_static_file(&attr_file, filename, "w") != CP40_OK) {
    return cp40_sys_error("cannot open attribute file for writing");
  }
  if (cp40_hash_write_comment(attr_file,
		       "Editable CP40 configuration file.") != CP40_OK) {
    goto error;
  }
  if (flags & CP40_ATTR_STANDARD) {
    for (i = 0; std_attr[i].name; i++) {
      if (print_attr(attr_file, instance, std_attr[i].name,
		     std_attr[i].flags) != CP40_OK) {
	goto error;
      }
    }
  } else if (flags & CP40_ATTR_PREPROCESSING) {
    for (i = 0; std_attr[i].name; i++) {
      if (std_attr[i].flags & PREPROCESSING) {
	if (print_attr(attr_file, instance, std_attr[i].name,
		      std_attr[i].flags) != CP40_OK) {
	  goto error;
	}
      }
    }
  }
  if (flags & CP40_ATTR_EXTRA) {
    int index = -1, status;
    const char *name, *value;
    while ((status= cp40_get_next_extra(instance, &index, &name,
					     &value)) == 1) {
      if (print_attr(attr_file, instance, name, EXTRA)) {
	goto error;
      }
    }
    if (status != 0) {
      goto error;
    }
  }
  if (cp40_hash_write_comment
            (attr_file,"End of editable CP40 configuration file.") != CP40_OK) {
    goto error;
  }
 done:
  cp40_close_static_file(&attr_file);
  return result;
 error:
  result= CP40_ERROR;
  goto done;
}

int cp40_restore_attr(cp40_instance_t *instance, const char *filename,
		      int flags)
{
  cp40_hash_table_t table;
  cp40_hash_entry_t *entry;
  cp40_hash_search_t search;
  int i, c, status;
  int result = CP40_OK;
  register const char *name, *std;
  register int n;

  if (filename == NULL || filename[0] == 0) {
    return cp40_error("bad attribute file name");
  }
  cp40_hash_init_table(&table, CP40_HASH_STRING_KEYS);
  if ((result= cp40_hash_restore(&table, filename)) != CP40_OK) {
    goto done;
  }
  for (entry=cp40_hash_first_entry(&table, &search);
       entry!=NULL;
       entry=cp40_hash_next_entry(&search)) {
    name= cp40_hash_get_key(entry);

    /* Search among standard attribute names. */
    status= 0;
    for (i=0 ; (std= std_attr[i].name) != NULL ; i++) {
      for (n=0 ; std[n] && std[n] == name[n] ; n++) ;
      if (n <= 0 || std[n] != 0) {
	continue;
      }
      if ((c = name[n]) == ',') {
	if ((unsigned int)(name[n + 1] - '0') >= 4 || name[n + 2] != 0) {
	  result = cp40_error("bad channel for attribute: %s", name);
	  goto done;
	}
      } else if (c != 0) {
	continue;
      }
      if (std_attr[i].flags & READ_ONLY) {
	cp40_warning("cannot set read-only attribute: %s", name);
	status = -1;
      } else if ((flags & CP40_ATTR_STANDARD)
		 || ((std_attr[i].flags & PREPROCESSING)
		     && (flags & CP40_ATTR_PREPROCESSING))) {
	status = 1;
      } else {
	status = 0;
      }
      break;
    }

    /* Set attribute if required. */
    if (status == 1 || (status == 0 && (flags & CP40_ATTR_EXTRA))) {
      result = cp40_set(instance, name, cp40_hash_get_value(entry));
      if (result != CP40_OK) {
	goto done;
      }
    }
  }

done:
  cp40_hash_delete_table(&table);
  return result;
}

static int print_attr(FILE *file, cp40_instance_t *instance, const char *name,
		      int flags)
{
  int chn, com= (flags&READ_ONLY)!=0;
  cp40_dstr_t dstr;

  if (flags == IGNORED) return CP40_OK;
  cp40_dstr_init(&dstr);
  if (! (flags & PER_CHANNEL)) {
    if (cp40_get(instance, name, &dstr) != CP40_OK ||
	cp40_hash_write_entry(file, name, cp40_dstr_value(&dstr), com)
	!= CP40_OK) {
      goto error;
    }
  } else {
    char fullname[32]; /* large enough for standard attribute + channel */
    if (strlen(name) > sizeof(fullname) - 3) {
      cp40_error("BUG: standard attribute name too long (%s:%d)",
	      __FILE__, __LINE__);
      goto error;
    }
    for EACH_CHANNEL(chn) {
      sprintf(fullname, "%s,%d", name, chn);
      if (cp40_get(instance, fullname, &dstr) != CP40_OK ||
	  cp40_hash_write_entry(file, fullname, cp40_dstr_value(&dstr), com)
	  != CP40_OK) {
	goto error;
      }
    }
  }
  cp40_dstr_free(&dstr);
  return CP40_OK;
 error:
  cp40_dstr_free(&dstr);
  return CP40_ERROR;
}

/*---------------------------------------------------------------------------*/

extern int cp40_copy_attr(cp40_instance_t *dst_ptr,
			  cp40_instance_t *src_ptr, int flags);
/*	Copy attributes between two CP40 data instances according to given
 *	flags.  Return standard result: CP40_OK or CP40_ERROR.*/

static int copy_attr(cp40_instance_t *dst_ptr, cp40_instance_t *src_ptr,
		     const char *name, int flags);
/*	Copy an attribute value pair (or one value per channel) according
	to given flags.  Return standard result: CP40_OK or CP40_ERROR.*/

int cp40_copy_attr(cp40_instance_t *dst_ptr,
	           /* const */ cp40_instance_t *src_ptr, int flags)
{
  long i;

  if (flags & CP40_ATTR_STANDARD) {
    for (i=0 ; std_attr[i].name!=NULL ; i++) {
      if (copy_attr(dst_ptr, src_ptr,
		    std_attr[i].name, std_attr[i].flags) != CP40_OK) {
	return CP40_ERROR;
      }
    }
  } else if (flags & CP40_ATTR_PREPROCESSING) {
    for (i=0 ; std_attr[i].name ; i++) {
      if (std_attr[i].flags & PREPROCESSING) {
	if (copy_attr(dst_ptr, src_ptr,
		      std_attr[i].name, std_attr[i].flags) != CP40_OK) {
	  return CP40_ERROR;
	}
      }
    }
  }

  if (flags & CP40_ATTR_EXTRA) {
    int index = -1, status;
    const char *name, *value;
    for (;;) {
      status= cp40_get_next_extra(src_ptr, &index, &name, &value);
      if (status==0) break;
      if (status!=1 || copy_attr(dst_ptr, src_ptr, name, EXTRA) != CP40_OK) {
	return CP40_ERROR;
      }
    }
  }
  return CP40_OK;
}

static int copy_attr(cp40_instance_t *dst_ptr,
		     /* const */ cp40_instance_t *src_ptr,
		     const char *name, int flags)
{
  cp40_dstr_t dstr;
  int chn;

  if (flags == IGNORED) return CP40_OK;
  cp40_dstr_init(&dstr);
  if (flags & PER_CHANNEL) {
    char fullname[32]; /* large enough for standard attribute + channel */
    for EACH_CHANNEL(chn) {
      sprintf(fullname, "%s,%d", name, chn);
      if (cp40_get(src_ptr, fullname, &dstr) != CP40_OK
	  || cp40_set(dst_ptr, fullname, cp40_dstr_value(&dstr)) != CP40_OK) {
	goto error;
      }
    }
  } else {
    if (cp40_get(src_ptr, name, &dstr) != CP40_OK
	|| cp40_set(dst_ptr, name, cp40_dstr_value(&dstr)) != CP40_OK) {
      goto error;
    }
  }
  cp40_dstr_free(&dstr);
  return CP40_OK;
 error:
  cp40_dstr_free(&dstr);
  return CP40_ERROR;
}

/*---------------------------------------------------------------------------*/

static int parse_int(const char *string, int *result);
static int parse_double(const char *string, double *result);
static int parse_doubles(const char *string, double *result, int number);

static int parse_channel(int *channel_ptr, const char *name);
/*	Parse channel number in attribute name in the form: "ATTRIBUTE,CHANNEL"
  	Check and store channel number, if any, in *CHANNEL_PTR; if none,
 	store ATTR_ALL_CHANNELS.  Return: CP40_OK or CP40_ERROR. In case of error,
	a message is left that can be recovered with cp40_error_info(). */

static int parse_channel(int *channel_ptr, const char *name)
{
  int c;

  while ((c= *name++) != 0) {
    if (c == ',') {
      if (name[1] != 0 || (unsigned int)(c= name[0]-'0') >= CP40_NCHANNELS) {
	return cp40_error("BUG: malformed \"attribute,channel\"");
      }
      *channel_ptr = c;
      return CP40_OK;
    }
  }
  *channel_ptr = ATTR_ALL_CHANNELS;
  return CP40_OK;
}

static int parse_int(const char *string, int *result)
{
  char *end;
  if (! string) goto bad_value;
  while (CP40_IS_SPACE(string[0])) string++;
  *result= strtol(string, &end, 0);
  if (end<=string) goto bad_value;
  while (CP40_IS_SPACE(end[0])) end++;
  if (*end) goto bad_value;
  return CP40_OK;
bad_value:
  return cp40_error("cannot convert string to integer value");
}

static int parse_double(const char *string, double *result)
{
  char *end;
  if (! string) goto bad_value;
  while (CP40_IS_SPACE(string[0])) string++;
  *result= strtod(string, &end);
  if (end<=string) goto bad_value;
  while (CP40_IS_SPACE(end[0])) end++;
  if (*end) goto bad_value;
  return CP40_OK;
bad_value:
  return cp40_error("cannot convert string to floating-point value");
}

#if 0 /* the following routine is not yet needed */
static int parse_ints(const char *string, int *result, int number);
static int parse_ints(const char *string, int *result, int number)
{
  int i;
  char *next= NULL, *cache;

  if (! (cache= clone_in_cache(string))) return CP40_ERROR;
  for (i=0 ; i<number ; i++) {
    if (parse_int(cp40_str_token(cache, &next, NULL, NULL), result+i) != CP40_OK) {
      return CP40_ERROR;
    }
  }
  if (cp40_str_token(cache, &next, NULL, NULL) != NULL) {
    return cp40_error("too many values");
  }
  return CP40_OK;
}
#endif

static int parse_doubles(const char *string, double *result, int number)
{
  int i;
  char *next= NULL, *cache;

  if (! (cache= clone_in_cache(string))) return CP40_ERROR;
  for (i=0 ; i<number ; i++) {
    if (parse_double(cp40_str_token(cache, &next, NULL, NULL), result+i) != CP40_OK) {
      return CP40_ERROR;
    }
  }
  if (cp40_str_token(cache, &next, NULL, NULL) != NULL) {
    return cp40_error("too many values");
  }
  return CP40_OK;
}

/*---------------------------------------------------------------------------*/

static int get_nphotons(cp40_instance_t *instance, const char *name,
			cp40_dstr_t *value);
/*	Get total number of photons for a given channel or all channels in
	CP40 data INSTANCE and store it as a human-readable string into VALUE.
	Return standard result: CP40_OK or CP40_ERROR (error message can be
	retrieved by cp40_error_info). */

static int get_nphotons(cp40_instance_t *instance, const char *name,
			cp40_dstr_t *value)
{
  int channel, number, nphotons[CP40_NCHANNELS];

  if (parse_channel(&channel, name) != CP40_OK) return CP40_ERROR;
  if (channel == ATTR_ALL_CHANNELS) {
    number= CP40_NCHANNELS;
    for EACH_CHANNEL(channel) {
      if ((nphotons[channel]= cp40_get_nphotons(instance, channel)) < 0) {
	return CP40_ERROR;
      }
    }
  } else {
    number= 1;
    if ((nphotons[0]= cp40_get_nphotons(instance, channel)) < 0) {
      return CP40_ERROR;
    }
  }
  return cp40_dstr_append_ints(value, nphotons, number, ' ');
}

/*---------------------------------------------------------------------------*/
/* SCALAR INTEGER ATTRIBUTES */

#undef FUNCTION
#define FUNCTION(NAME, FUNC)						      \
static int NAME(cp40_instance_t *instance, const char *name, cp40_dstr_t *value) \
{ return cp40_dstr_append_int(value, FUNC(instance), 0); }
FUNCTION(get_exposure,      cp40_get_exposure)
FUNCTION(get_frame_count,   cp40_get_frame_count)
FUNCTION(get_frame_date,    cp40_get_frame_date)
FUNCTION(get_nframes,       cp40_get_nframes)
FUNCTION(get_nstaturations, cp40_get_nsaturations)

#undef FUNCTION
#define FUNCTION(NAME, FUNC)					\
static int NAME(cp40_instance_t *instance, const char *name,	\
		const char *value)				\
{								\
  int val;							\
  if (parse_int(value, &val) != CP40_OK) return CP40_ERROR;	\
  return FUNC(instance, val);					\
}
FUNCTION(set_exposure, cp40_set_exposure)

/*---------------------------------------------------------------------------*/
/* OPERATIONS INVOLVING A FLOATING-POINT SCALAR */

#undef FUNCTION
#define FUNCTION(NAME, FUNC)						      \
static int NAME(cp40_instance_t *instance, const char *name, cp40_dstr_t *value) \
{ return cp40_dstr_append_double(value, FUNC(instance), 0); }
FUNCTION(get_angle, cp40_get_angle)
FUNCTION(get_scale, cp40_get_scale)

#undef FUNCTION
#define FUNCTION(NAME, FUNC)					\
static int NAME(cp40_instance_t *instance, const char *name,	\
		const char *value)				\
{								\
  double val;							\
  if (parse_double(value, &val) != CP40_OK) return CP40_ERROR;	\
  return FUNC(instance, val);					\
}
FUNCTION(set_angle, cp40_set_angle)
FUNCTION(set_scale, cp40_set_scale)

/*---------------------------------------------------------------------------*/
/* KEYWORD OPERATIONS */

static int get_keyword(const keyword_info_t *keyword,
		       int ival, cp40_dstr_t *value);
static int parse_keyword(int *int_ptr, const keyword_info_t *keyword,
			 const char *value);

static int get_keyword(const keyword_info_t *keyword,
		       int ival, cp40_dstr_t *value)
{
  int i;

  for (i=0 ; keyword[i].sval!=NULL ; i++) {
    if (keyword[i].ival == ival) {
      return cp40_dstr_append(value, keyword[i].sval, -1);
    }
  }
  return cp40_error("unexpected keyword integer value");
}

static int parse_keyword(int *int_ptr, const keyword_info_t *keyword,
			 const char *value)
{
  int i;

  if ((value != NULL) && (value[0] != 0)) {
    for (i=0 ; keyword[i].sval!=NULL ; i++) {
      if (strcmp(keyword[i].sval, value) == 0) {
	*int_ptr= keyword[i].ival;
	return CP40_OK;
      }
    }
  }
  return cp40_error("bad keyword value");
}

#undef FUNCTION
#define FUNCTION(NAME, KEYWORD, FUNC)					      \
static int NAME(cp40_instance_t *instance, const char *name, cp40_dstr_t *value) \
{  return get_keyword(KEYWORD, FUNC(instance), value); }
FUNCTION(get_correction, keyword_correction, cp40_get_correction)
FUNCTION(get_parity,     keyword_parity,     cp40_get_parity)
FUNCTION(get_pixel,      keyword_pixel,      cp40_get_pixel)
FUNCTION(get_spurious,   keyword_spurious,   cp40_get_spurious)

#undef FUNCTION
#define FUNCTION(NAME, KEYWORD, FUNC) 					\
static int NAME(cp40_instance_t *instance,				\
		const char *name, const char *value)			\
{									\
  int ival;								\
  if (parse_keyword(&ival, KEYWORD, value) != CP40_OK) return CP40_ERROR;	\
  return FUNC(instance, ival);						\
}
FUNCTION(set_correction, keyword_correction, cp40_set_correction)
FUNCTION(set_parity,     keyword_parity,     cp40_set_parity)
FUNCTION(set_pixel,      keyword_pixel,      cp40_set_pixel)
FUNCTION(set_spurious,   keyword_spurious,   cp40_set_spurious)

/*---------------------------------------------------------------------------*/
/* RANGE OPERATIONS */

static int append_range_value(cp40_dstr_t *result, double value, int wild);

typedef int get_range_op_t(cp40_instance_t *instance, int channel,
			   cp40_range_t *range);

typedef int set_range_op_t(cp40_instance_t *instance, int channel,
			   cp40_range_t *range);

static int append_range_value(cp40_dstr_t *result, double value, int wild)
{
  if (wild) {
    if (cp40_dstr_length(result)) return cp40_dstr_append(result, " *", 2);
    return cp40_dstr_append_char(result, '*');
  }
  return cp40_dstr_append_double(result, value, ' ');
}

static int get_range(cp40_instance_t *instance,
		     const char *name, cp40_dstr_t *value)
{
  cp40_range_t range;
  int channel, flag;
  get_range_op_t *func;

  if (parse_channel(&flag, name) != CP40_OK) return CP40_ERROR;
  switch (name[0]) {
  case 'd': func= cp40_get_distance; break;
  case 'f': func= cp40_get_flux;     break;
  case 'r': func= cp40_get_ratio;    break;
  case 's': func= cp40_get_spread;   break;
  default: return cp40_error("(BUG) unexpected range attribute");
  }

  channel= (flag == ATTR_ALL_CHANNELS) ? 0 : flag;
  for (;;) {
    if ((*func)(instance, channel, &range) != CP40_OK
	|| append_range_value(value, range.min,
			      !(range.flags&CP40_MIN_SET)) != CP40_OK
	|| append_range_value(value, range.max,
			      !(range.flags&CP40_MAX_SET)) != CP40_OK) {
      return CP40_ERROR;
    }
    if (flag != ATTR_ALL_CHANNELS || ++channel >= CP40_NCHANNELS) break;
  }
  return CP40_OK;
}

static int set_range(cp40_instance_t *instance, const char *name,
		     const char *value)
{
  char *next, *token, *cache;
  cp40_range_t range;
  int flag, channel, count;
  set_range_op_t *func;

  /* Figure out which function to use to set range. */
  switch (name[0]) {
  case 'd': func= cp40_set_distance; break;
  case 'f': func= cp40_set_flux;     break;
  case 'r': func= cp40_set_ratio;    break;
  case 's': func= cp40_set_spread;   break;
  default: return cp40_error("(BUG) unexpected range attribute");
  }

  /* Get channel number. */
  if (parse_channel(&flag, name) != CP40_OK) return CP40_ERROR;

  /* Parse string value. */
  if (! (cache= clone_in_cache(value))) return CP40_ERROR;
  for (count=0, next=NULL ; ; ) {
    range.flags= 0;
    if (! (token= cp40_str_token(cache, &next, NULL, NULL))) break;
    if (token[0]=='*') {
      if (token[1]!=0) goto bad_value;
    } else {
      if (parse_double(token, &range.min) != CP40_OK) goto bad_value;
      range.flags|= CP40_MIN_SET;
    }
    if (! (token= cp40_str_token(cache, &next, NULL, NULL))) goto bad_value;
    if (token[0]=='*') {
      if (token[1]!=0) goto bad_value;
    } else {
      if (parse_double(token, &range.max) != CP40_OK) goto bad_value;
      range.flags|= CP40_MAX_SET;
    }
    count+= 1;
    if (flag == ATTR_ALL_CHANNELS) {
      if ((channel= count-1) >= CP40_NCHANNELS) goto bad_value;
      if ((*func)(instance, channel, &range) != CP40_OK) return CP40_ERROR;
    } else {
      if (count!=1) goto bad_value;
      if ((*func)(instance, flag, &range) != CP40_OK) return CP40_ERROR;
    }
  }
  if (count!=4) {
    if (count!=1) goto bad_value;
    if (flag == ATTR_ALL_CHANNELS) {
      for (channel=1 ; channel<CP40_NCHANNELS ; channel++)
	if ((*func) (instance, channel, &range) != CP40_OK) return CP40_ERROR;
    }
  }
  return CP40_OK;
bad_value:
  return cp40_error("bad range value");
}

/*---------------------------------------------------------------------------*/
/* WINDOW OPERATIONS */

static int append_window_value(cp40_dstr_t *result, int value);
static int parse_window_value(const char *token, int *value_ptr);
typedef int get_window_op_t(cp40_instance_t *instance, int channel,
			    cp40_window_t *window_ptr);

static int get_window(cp40_instance_t *instance,
		      const char *name, cp40_dstr_t *value)
{
  cp40_window_t window;
  int channel, flag;
  get_window_op_t *func;

  if (parse_channel(&flag, name) != CP40_OK) return CP40_ERROR;
  switch (name[0]) {
  case 'e': func= cp40_get_extreme; break;
  case 'w': func= cp40_get_window; break;
  default: return cp40_error("(BUG) unexpected window attribute");
  }

  channel= flag == ATTR_ALL_CHANNELS ? 0 : flag;
  for (;;) {
    if ((*func)(instance, channel, &window) != CP40_OK
	|| append_window_value(value, window.xmin) != CP40_OK
	|| append_window_value(value, window.xmax) != CP40_OK
	|| append_window_value(value, window.ymin) != CP40_OK
	|| append_window_value(value, window.ymax) != CP40_OK) {
      return CP40_ERROR;
    }
    if (flag != ATTR_ALL_CHANNELS || ++channel >= CP40_NCHANNELS) break;
  }
  return CP40_OK;
}

static int set_window(cp40_instance_t *instance, const char *name,
		      const char *value)
{
  char *next, *token, *cache;
  cp40_window_t window;
  int flag, channel, count;

  /* Get channel number. */
  if (parse_channel(&flag, name) != CP40_OK) return CP40_ERROR;

  /* Parse string value. */
  if (! (cache= clone_in_cache(value))) return CP40_ERROR;
  for (count=0, next=NULL ; ; ) {
    /* Parse one window settings. */
    if (! (token= cp40_str_token(cache, &next, NULL, NULL))) break;
    if (parse_window_value(token, &window.xmin) != CP40_OK
	|| ! (token= cp40_str_token(cache, &next, NULL, NULL))
	|| parse_window_value(token, &window.xmax) != CP40_OK
	|| ! (token= cp40_str_token(cache, &next, NULL, NULL))
	|| parse_window_value(token, &window.ymin) != CP40_OK
	|| ! (token= cp40_str_token(cache, &next, NULL, NULL))
	|| parse_window_value(token, &window.ymax) != CP40_OK) {
      goto bad_value;
    }

    /* Set the corresponding window attribute. */
    count+= 1;
    if (flag == ATTR_ALL_CHANNELS) {
      if ((channel= count-1) >= CP40_NCHANNELS) goto bad_value;
      if (cp40_set_window(instance, channel, &window) != CP40_OK) return CP40_ERROR;
    } else {
      if (count!=1) goto bad_value;
      if (cp40_set_window(instance, flag, &window) != CP40_OK) return CP40_ERROR;
    }
  }
  if (count!=4) {
    if (count!=1) goto bad_value;
    if (flag == ATTR_ALL_CHANNELS) {
      for (channel=1 ; channel<CP40_NCHANNELS ; channel++) {
	if (cp40_set_window(instance, channel, &window) != CP40_OK) {
	  return CP40_ERROR;
	}
      }
    }
  }
  return CP40_OK;
bad_value:
  return cp40_error("bad window value");
}

static int append_window_value(cp40_dstr_t *result, int value)
{
  if (value==CP40_WINDOW_WILD) {
    if (cp40_dstr_length(result)) return cp40_dstr_append(result, " *", 2);
    return cp40_dstr_append_char(result, '*');
  }
  return cp40_dstr_append_int(result, value, ' ');
}

static int parse_window_value(const char *token, int *value_ptr)
{
  if (token[0]=='*') {
    if (token[1]!=0) return CP40_ERROR;
    *value_ptr= CP40_WINDOW_WILD;
    return CP40_OK;
  }
  return parse_int(token, value_ptr);
}

/*---------------------------------------------------------------------------*/
/* DISTORTION COEFFICIENTS */

static int get_coefs(cp40_instance_t *instance,
		     const char *name, cp40_dstr_t *value)
{
  double coefs[CP40_COEF_COUNT];

  if (name[0] == 'r') cp40_get_raw_coefs(instance, coefs);
  else cp40_get_coefs(instance, coefs);
  return cp40_dstr_append_doubles(value, coefs, CP40_COEF_COUNT, ' ');
}

static int set_coefs(cp40_instance_t *instance, const char *name,
		     const char *value)
{
  double coefs[CP40_COEF_COUNT];

  if (parse_doubles(value, coefs, CP40_COEF_COUNT) != CP40_OK) {
    return cp40_error("bad list of coefficients values");
  }
  if (cp40_set_raw_coefs(instance, coefs) != CP40_OK) {
    return CP40_ERROR;
  }
  return CP40_OK;
}

/*_____________________________________________________________________________
 *
 * set_read_only --
 *
 *	Set read-only attribute value.
 *
 * Return:
 *	Always an error: CP40_ERROR.
 *
 * Side effects:
 *	See description..
 *_____________________________________________________________________________
 */

static int set_read_only(cp40_instance_t *instance, const char *name,
			 const char *value)
{
  return cp40_error("cannot set read-only attribute \"%s\"", name);
}

/*---------------------------------------------------------------------------*/

static int get_bogus(cp40_instance_t *instance,
		     const char *name, cp40_dstr_t *value)
{ return cp40_error("cannot get attribute value (bad initialization)"); }

static int set_bogus(cp40_instance_t *instance, const char *name,
		     const char *value)
{ return cp40_error("cannot store attribute value (bad initialization)"); }

/*---------------------------------------------------------------------------*/
/* EXTRA (NON-STANDARD) ATTRIBUTES */

static int attr_extra_index(cp40_instance_t *instance, const char *name,
			    int *new_ptr, int report_error);
/*	Given extra (i.e. non standard) attribute name, get its index in
	the attribute list.  If NEW_PTR is non-NULL, use cp40_hash_create_entry
	instead of cp40_hash_find_entry and augment extra attribute list as
	needed.  If REPORT_ERROR is non-zero and NEW_PTR is NULL, store an
	error message if attribute is not found.  Return positive index or
	-1 in case of error. */

int cp40_get_next_extra(cp40_instance_t *instance, int *index_ptr,
			const char **name_ptr, const char **value_ptr)
{
  int index;

  *value_ptr= NULL;
  *name_ptr= NULL;
  index= *index_ptr;
  if (index < -1) {
    index= -1;
  }
  for (index+=1 ; index < instance->extra_count ; index++) {
    if (instance->extra_list[index].string != NULL) {
      if (instance->extra_list[index].entry == NULL) {
	cp40_error("BUG: malformed extra attribute list");
	return -1;
      }
      *index_ptr= index;
      *value_ptr= instance->extra_list[index].string;
      *name_ptr= cp40_hash_get_key(instance->extra_list[index].entry);
      return 1;
    }
  }
  *index_ptr= index;
  return 0;
}

static int get_extra(cp40_instance_t *instance,
		     const char *name, cp40_dstr_t *value)
{
  int index;

  if ((index= attr_extra_index(instance, name, NULL, 1)) < 0) {
    return CP40_ERROR;
  }
  return cp40_dstr_append(value, instance->extra_list[index].string, -1);
}

static int set_extra(cp40_instance_t *instance,
		     const char *name, const char *value)
{
  cp40_extra_t *extra_list;
  const char *old_string;
  int new, index;

  if (value == NULL) {
    /* Delete attribute entry. */
    index= attr_extra_index(instance, name, NULL, 0);
    if (index >= 0) {
      extra_list= instance->extra_list + index;
      cp40_hash_delete_entry(extra_list->entry);
      extra_list->entry= NULL;
      cp40_mem_free((void *)extra_list->string);
      extra_list->string= NULL;
      if (index == instance->extra_count - 1) {
	instance->extra_count-= 1;
      }
    }
    return CP40_OK;
  }

  /* (Re)set attribute value. */
  index= attr_extra_index(instance, name, &new, 1);
  if (index < 0) {
    return CP40_ERROR;
  }
  extra_list= instance->extra_list + index;
  old_string= new ? extra_list->string : NULL;
  extra_list->string= cp40_str_clone(value);
  if (extra_list->string == NULL) {
    if (new) {
      /* Delete new item. */
      cp40_hash_delete_entry(extra_list->entry);
      extra_list->entry= NULL;
      instance->extra_count-= 1;	/* New, so must be last item. */
    } else {
      /* Restore old value. */
      extra_list->string= old_string;
    }
    return cp40_error("cannot store attribute value");
  }
  if (old_string != NULL) {
    cp40_mem_free((void *)old_string);
  }
  return CP40_OK;
}

static int attr_extra_index(cp40_instance_t *instance, const char *name,
			    int *new_ptr, int report_error)
{
  cp40_hash_entry_t *entry;
  cp40_extra_t *extra_list;
  int index;

  if (name == NULL) {
    if (report_error) {
      cp40_error("bad attribute name (NULL string)");
    }
    return -1;
  }
  if (new_ptr != NULL) {
    entry= cp40_hash_create_entry(&instance->extra_table, name, new_ptr);
    if (entry == NULL) {
      cp40_error("unable to create attribute entry");
      return -1;
    }
  } else {
    entry= cp40_hash_find_entry(&instance->extra_table, name);
    if (entry == NULL) {
      if (report_error) {
	cp40_error("no such attribute");
      }
      return -1;
    }
  }
  if (new_ptr == NULL || *new_ptr == 0) {
    /* Already existing entry, return index stored has hash value. */
    index = (int)cp40_hash_get_value(entry);
    if (index < 0 || index >= instance->extra_count
	|| (instance->extra_list)[index].string == NULL
	|| (instance->extra_list)[index].entry != entry) {
      cp40_error("BUG: malformed extra attribute list");
      return -1;
    }
    return index;
  }

  /* New entry. */
#undef EXTRA_CHUNK
#define EXTRA_CHUNK 50
  if (instance->extra_count < instance->extra_size) {
    extra_list = instance->extra_list;
  } else {
    /* Allocate/reallocate extra attribute list by bucket of
       EXTRA_CHUNK items. */
    if (instance->extra_list == NULL) {
      extra_list = malloc(sizeof(*extra_list)*EXTRA_CHUNK);
    } else {
      extra_list = realloc(instance->extra_list,
			   sizeof(*extra_list)*(EXTRA_CHUNK +
						instance->extra_size));
    }
    if (extra_list == NULL) {
      cp40_hash_delete_entry(entry);
      cp40_sys_error("cannot (re)alloc extra-attribute list");
      return -1;
    }
    instance->extra_size+= EXTRA_CHUNK;
    instance->extra_list= extra_list;
    for (index = instance->extra_count;
	 index < instance->extra_size; index++) {
      extra_list[index].entry= NULL;
      extra_list[index].string= NULL;
    }
  }
  index = instance->extra_count;
  extra_list[index].entry = entry;
  extra_list[index].string = NULL;
  ++instance->extra_count;
  cp40_hash_set_value(entry, (void *)index, CP40_HASH_STATIC);
  return index;
}

/*---------------------------------------------------------------------------*/
